import logging

from PySide import QtGui

from spreads.plugin import HookPlugin
import gui
import gui_rc

logger = logging.getLogger('spreadsplug.gui')


class GuiCommand(HookPlugin):
    @classmethod
    def add_command_parser(cls, rootparser):
        guiparser = rootparser.add_parser(
            'gui', help="Start the GUI wizard")
        guiparser.set_defaults(subcommand=GuiCommand.wizard)

    @staticmethod
    def wizard(config):
        logger.debug("Starting GUI")
        app = QtGui.QApplication([])

        wizard = gui.SpreadsWizard(config)
        wizard.show()
        app.exec_()



# -*- coding: utf-8 -*-

from __future__ import division, unicode_literals

import logging
import os
import subprocess

from spreads.plugin import HookPlugin
from spreads.util import MissingDependencyException, find_in_path

if not find_in_path('pdfbeads'):
    raise MissingDependencyException("Could not find executable `pdfbeads` in"
                                     " $PATH. Please install the appropriate"
                                     " package(s)!")

logger = logging.getLogger('spreadsplug.pdfbeads')


class PDFBeadsPlugin(HookPlugin):
    __name__ = 'pdfbeads'

    def output(self, path):
        logger.info("Assembling PDF.")
        img_dir = path / 'done'
        pdf_file = path / 'out' / "{0}.pdf".format(path.name)
        img_files = [unicode(x.name) for x in sorted(img_dir.glob('*.tif'))]
        cmd = ["pdfbeads", "-d"] + img_files + ["-o", unicode(pdf_file)]
        logger.debug("Running " + " ".join(cmd))
        # NOTE: pdfbeads only finds *html files for the text layer in the
        #       working directory...
        os.chdir(unicode(img_dir))
        output = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        logger.debug("Output:\n{0}".format(output))



import shutil
import tempfile
import unittest

import spreads.vendor.confit as confit
from mock import call, MagicMock as Mock, patch
from spreads.vendor.pathlib import Path

import spreads.util
spreads.util.find_in_path = Mock(return_value=True)
import spreadsplug.autorotate as autorotate


class TestAutorotate(unittest.TestCase):
    def setUp(self):
        self.config = confit.Configuration('test_autorotate')
        self.config['autorotate']['rotate_odd'] = -90
        self.config['autorotate']['rotate_even'] = 90
        self.path = Path(tempfile.mkdtemp())
        (self.path / 'raw').mkdir()

    def tearDown(self):
        shutil.rmtree(unicode(self.path))

    def test_process(self):
        test_files = [self.path / 'raw' / x
                      for x in ('000.jpg', '001.jpg', '002.jpg')]
        map(lambda x: x.touch(), test_files)

        mock_pool = Mock()
        autorotate.futures.ProcessPoolExecutor = Mock(return_value=mock_pool)
        with patch('spreadsplug.autorotate._get_exif_orientation') as geo:
            geo.side_effect = (6, 8, -1)
            plug = autorotate.AutoRotatePlugin(self.config)
            plug.process(self.path)

        assert geo.call_count == 3
        mock_pool.__enter__().submit.assert_has_calls([
            call(autorotate.rotate_image, test_files[0], rotation=90),
            call(autorotate.rotate_image, test_files[1], rotation=-90)
        ])

    def test_process_inverse(self):
        test_files = [self.path / 'raw' / '000.jpg']
        map(lambda x: x.touch(), test_files)
        mock_pool = Mock()
        autorotate.futures.ProcessPoolExecutor = Mock(return_value=mock_pool)
        self.config['autorotate']['rotate_odd'] = 90
        self.config['autorotate']['rotate_even'] = -90
        plug = autorotate.AutoRotatePlugin(self.config)
        with patch('spreadsplug.autorotate._get_exif_orientation') as geo:
            geo.return_value = 6
            plug.process(self.path)
        mock_pool.__enter__().submit.assert_called_with(
            autorotate.rotate_image, test_files[0], rotation=-90
        )

    def test_rotate_image(self):
        mock_img = Mock()
        mock_img.__enter__().height = 200
        mock_img.__enter__().width = 800
        autorotate.wand.image.Image = Mock(return_value=mock_img)
        with patch('spreadsplug.autorotate.subprocess.check_output'):
            autorotate.rotate_image('foo.jpg', 90)
        assert mock_img.__enter__().rotate.call_args_list == [call(90)]
        assert mock_img.__enter__().save.call_args_list == [
            call(filename='foo.jpg')]



import unittest
from itertools import chain, repeat

import spreads.vendor.confit as confit
from mock import MagicMock as Mock, patch
from spreads.vendor.pathlib import Path

import spreads.plugin as plugin

with patch('subprocess.check_output') as mock_co:
    mock_co.return_value = "".join(chain(
        repeat("\n", 7),
        ("scantailor-cli [options] <images|directory|-> <output>",))
    )
    import spreadsplug.scantailor as scantailor


class TestScanTailor(unittest.TestCase):
    @patch('subprocess.check_output')
    def setUp(self, mock_co):
        self.config = confit.Configuration('test_scantailor')

        tmpl = scantailor.ScanTailorPlugin.configuration_template()
        for key, option in tmpl.items():
            if option.selectable:
                self.config['scantailor'][key] = option.value[0]
            else:
                self.config['scantailor'][key] = option.value
        self.stplug = scantailor.ScanTailorPlugin(self.config)

    def tearDown(self):
        plugin.SpreadsNamedExtensionManager._instance = None

    def test_generate_configuration(self):
        scantailor.subprocess.call = Mock()
        assert self.stplug._enhanced
        # TODO: Setup up some config variables
        self.stplug._generate_configuration(Path('/tmp/foo.st'),
                                            Path('/tmp/raw'),
                                            Path('/tmp/out'))
        # TODO: Check the sp.call for the correct parameters

    def test_generate_configuration_noenhanced(self):
        scantailor.subprocess.call = Mock()
        # TODO: Setup up some config variables
        self.stplug._enhanced = False
        imgdir_mock = Mock(wraps=Path('/tmp/raw'))
        mock_imgs = [imgdir_mock/"foo.jpg", imgdir_mock/"bar.jpg"]
        imgdir_mock.iterdir.return_value = mock_imgs
        self.stplug._generate_configuration(Path('/tmp/foo.st'), imgdir_mock,
                                            Path('/tmp/out'))
        assert (unicode(mock_imgs[0])
                in scantailor.subprocess.call.call_args[0][0])

    def test_split_configuration(self):
        # TODO: Provide a sample configuration
        # TODO: Check for number of files
        # TODO: Check if the files have all the pages
        pass

    def test_generate_output(self):
        scantailor.subprocess.Popen = Mock()
        self.stplug._split_configuration = Mock()
        self.stplug._generate_output('/tmp/foo.st', '/tmp')

    def test_process(self):
        scantailor.subprocess.call = Mock()
        self.stplug._generate_configuration = Mock()
        self.stplug._generate_output = Mock()
        self.stplug.config['autopilot'] = True
        self.stplug.process(Path('/tmp'))
        assert scantailor.subprocess.call.call_count == 0
        self.stplug.config['autopilot'] = False
        self.stplug.process(Path('/tmp'))
        scantailor.subprocess.call.assert_called_with(
            ['scantailor', '/tmp/tmp.ScanTailor'])



# -*- coding: utf-8 -*-

# Copyright (c) 2013 Johannes Baiter. All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
spreads CLI code.
"""

from __future__ import division, unicode_literals, print_function

import argparse
import logging
import os
import sys
import time

import colorama
import pkg_resources
import spreads.vendor.confit as confit

from spreads.workflow import Workflow
from spreads.plugin import (get_driver, get_devices, get_pluginmanager,
                            setup_plugin_config, get_relevant_extensions,
                            DeviceFeatures)
from spreads.util import DeviceException, ColourStreamHandler

# Kudos to http://stackoverflow.com/a/1394994/487903
try:
    from msvcrt import getch
except ImportError:
    def getch():
        """ Wait for keypress on stdin.

        :returns: unicode -- Value of character that was pressed

        """
        import tty
        import termios
        fd = sys.stdin.fileno()
        old = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            return sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old)


def colorize(text, color):
    return color + text + colorama.Fore.RESET


def _select_driver():
    print(colorize("Please select a device driver from the following list:",
                   colorama.Fore.BLUE))
    available_drivers = list(
        pkg_resources.iter_entry_points('spreadsplug.devices'))
    for pos, ext in enumerate(available_drivers, 1):
        print("  [{0}]: {1}".format(pos, ext.name))
    while True:
        selection = raw_input("Select a driver: ")
        if not selection.isdigit() or int(selection) > len(available_drivers):
            print(colorize("Please select a number in the range of 1 to {0}"
                           .format(len(available_drivers)), colorama.Fore.RED))
            continue
        driver = unicode(available_drivers[int(selection)-1].name)
        print(colorize("Selected \"{0}\" as device driver".format(driver),
                       colorama.Fore.GREEN))
        return driver


def _select_plugins(selected_plugins=None):
    if selected_plugins is None:
        selected_plugins = []
    print("Please select your desired plugins from the following list:")
    available_plugins = list(
        pkg_resources.iter_entry_points('spreadsplug.hooks'))
    while True:
        for pos, ext in enumerate(available_plugins, 1):
            print("  {0} {1}: {2}"
                  .format('x' if ext.name in selected_plugins else ' ',
                          pos, ext.name))
        selection = raw_input("Select a plugin (or hit enter to finish): ")
        if not selection:
            break
        if not selection.isdigit() or int(selection) > len(available_plugins):
            print(colorize("Please select a number in the range of 1 to {0}"
                           .format(len(available_plugins)), colorama.Fore.RED))
            continue
        plugin_name = available_plugins[int(selection)-1].name
        if plugin_name in selected_plugins:
            selected_plugins.remove(plugin_name)
        else:
            selected_plugins.append(plugin_name)
    return selected_plugins


def _setup_processing_pipeline(config):
    pm = get_pluginmanager(config)
    extensions = [ext.name for ext in get_relevant_extensions(pm, ['process'])]
    print("The following postprocessing plugins were detected:")
    print("\n".join(" - {0}".format(ext) for ext in extensions))
    while True:
        answer = raw_input("Please enter the extensions in the order that they"
                           " should be invoked, separated by commas:\n")
        plugins = [x.strip() for x in answer.split(',')]
        if any(x not in extensions for x in plugins):
            print(colorize("At least one of the entered extensions was not"
                           "found, please try again!", colorama.Fore.RED))
        else:
            break
    config["plugins"] = plugins + [x for x in config["plugins"].get()
                                   if x not in plugins]


def _set_device_target_page(config, target_page):
    print("Please connect and turn on the device labeled \'{0}\'"
          .format(target_page))
    print("Press any key when ready.")
    getch()
    devs = get_devices(config)
    if len(devs) > 1:
        raise DeviceException("Please ensure that only one device is"
                              " turned on!")
    if not devs:
        raise DeviceException("No device found!")
    devs[0].set_target_page(target_page)
    print(colorize("Configured \'{0}\' device.".format(target_page),
                   colorama.Fore.GREEN))
    print("Please turn off the device.")
    print("Press any key when ready.")
    getch()


def configure(config):
    config["driver"] = _select_driver()
    config["plugins"] = _select_plugins(config["plugins"].get())

    setup_plugin_config(config)
    _setup_processing_pipeline(config)

    cfg_path = os.path.join(config.config_dir(), confit.CONFIG_FILENAME)
    setup_plugin_config(config)

    driver = get_driver(config["driver"].get()).driver

    # We only need to set the device target_page if the driver supports
    # shooting with two devices
    if DeviceFeatures.IS_CAMERA in driver.features:
        answer = raw_input(
            "Do you want to configure the target_page of your devices?\n"
            "(Required for shooting with two devices) [y/n]: ")
        answer = True if answer.lower() == 'y' else False
        if answer:
            print("Setting target page on cameras")
            for target_page in ('odd', 'even'):
                _set_device_target_page(config, target_page)

        answer = raw_input("Do you want to setup the focus for your cameras? "
                           "[y/n]: ")
        answer = True if answer.lower() == 'y' else False
        if answer:
            print("Please turn on one of your capture devices.\n"
                  "Press any key to continue")
            getch()
            devs = get_devices(config)
            print("Please put a book with as little whitespace as possible"
                  "under your cameras.\nPress any button to continue")
            getch()
            focus = devs[0]._acquire_focus()
            config['device']['focus_distance'] = focus
    print("Writing configuration file to '{0}'".format(cfg_path))
    config.dump(filename=cfg_path)


def capture(config):
    path = config['path'].get()
    workflow = Workflow(config=config, path=path)
    if len(workflow.devices) != 2:
        raise DeviceException("Please connect and turn on two"
                              " pre-configured devices! ({0} were"
                              " found)".format(len(workflow.devices)))
    print(colorize("Found {0} devices!".format(len(workflow.devices)),
                   colorama.Fore.GREEN))
    if any(not x.target_page for x in workflow.devices):
        raise DeviceException("At least one of the devices has not been"
                              " properly configured, please re-run the"
                              " program with the \'configure\' option!")
    # Set up for capturing
    print("Setting up devices for capturing.")
    workflow.prepare_capture()
    # Start capture loop
    shot_count = 0
    pages_per_hour = 0
    capture_keys = workflow.config['capture']['capture_keys'].as_str_seq()
    print("({0}) capture | (r) retake last shot | (f) finish "
          .format("/".join(capture_keys)))
    while True:
        retake = False
        char = getch().lower()
        if char == 'f':
            break
        elif char == 'r':
            retake = True
        elif char not in capture_keys:
            continue
        workflow.capture(retake=retake)
        shot_count += len(workflow.devices)
        pages_per_hour = (3600/(time.time() -
                          workflow.capture_start))*shot_count
        status = ("\rShot {0: >3} pages [{1: >4.0f}/h] "
                  .format(unicode(shot_count), pages_per_hour))
        sys.stdout.write(status)
        sys.stdout.flush()
    workflow.finish_capture()
    if workflow.capture_start is None:
        return
    sys.stdout.write("\rShot {0} pages in {1:.1f} minutes, average speed was"
                     " {2:.0f} pages per hour\n"
                     .format(colorize(str(shot_count), colorama.Fore.GREEN),
                             (time.time() - workflow.capture_start)/60,
                             pages_per_hour))
    sys.stdout.flush()


def postprocess(config):
    path = config['path'].get()
    workflow = Workflow(config=config, path=path)
    workflow.process()


def output(config):
    path = config['path'].get()
    workflow = Workflow(config=config, path=path)
    workflow.output()


def wizard(config):
    # TODO: Think about how we can make this more dynamic, i.e. get list of
    #       options for plugin with a description for each entry
    print("==========================\n",
          "Starting capturing process\n",
          "==========================")
    capture(config)

    print("=======================\n"
          "Starting postprocessing\n"
          "=======================")
    postprocess(config)

    print("=================\n",
          "Generating output\n"
          "=================")
    output(config)


def setup_parser(config):
    def _add_argument_from_option(extname, key, option, parser):
        flag = "--{0}".format(key.replace('_', '-'))
        default = (option.value
                   if not option.selectable
                   else option.value[0])
        kwargs = {'help': ("{0} [default: {1}]"
                           .format(option.docstring, default)),
                  'dest': "{0}.{1}".format(extname, key)}
        if isinstance(option.value, basestring):
            kwargs['type'] = unicode
            kwargs['metavar'] = "<str>"
        elif isinstance(option.value, bool):
            kwargs['help'] = option.docstring
            if option.value:
                flag = "--no-{0}".format(key.replace('_', '-'))
                kwargs['help'] = ("Disable {0}"
                                  .format(option.docstring.lower()))
                kwargs['action'] = "store_false"
            else:
                kwargs['action'] = "store_true"
        elif isinstance(option.value, float):
            kwargs['type'] = float
            kwargs['metavar'] = "<float>"
        elif isinstance(option.value, int):
            kwargs['type'] = int
            kwargs['metavar'] = "<int>"
        elif option.selectable:
            kwargs['type'] = type(option.value[0])
            kwargs['metavar'] = "<{0}>".format("/".join(option.value))
            kwargs['choices'] = option.value
        else:
            raise TypeError("Unsupported option type")
        parser.add_argument(flag, **kwargs)

    def _add_device_arguments(name, parser):
        tmpl = get_driver(config["driver"]
                          .get()).driver.configuration_template()
        if not tmpl:
            return
        for key, option in tmpl.iteritems():
            try:
                _add_argument_from_option('device', key, option, parser)
            except:
                return

    def _add_plugin_arguments(hooks, parser):
        extensions = get_relevant_extensions(pluginmanager, hooks)
        for ext in extensions:
            tmpl = ext.plugin.configuration_template()
            if not tmpl:
                continue
            for key, option in tmpl.iteritems():
                try:
                    _add_argument_from_option(ext.name, key, option, parser)
                except:
                    continue

    pluginmanager = get_pluginmanager(config)
    rootparser = argparse.ArgumentParser(
        description="Scanning Tool for  DIY Book Scanner")
    subparsers = rootparser.add_subparsers()

    rootparser.add_argument(
        '--verbose', '-v', dest="loglevel", action="store_const",
        const="debug")

    wizard_parser = subparsers.add_parser(
        'wizard', help="Interactive mode")
    wizard_parser.add_argument(
        "path", type=unicode, help="Project path")
    wizard_parser.set_defaults(subcommand=wizard)

    config_parser = subparsers.add_parser(
        'configure', help="Perform initial configuration")
    config_parser.set_defaults(subcommand=configure)

    capture_parser = subparsers.add_parser(
        'capture', help="Start the capturing workflow")
    capture_parser.add_argument(
        "path", type=unicode, help="Project path")
    capture_parser.set_defaults(subcommand=capture)
    # Add arguments from plugins
    for parser in (capture_parser, wizard_parser):
        _add_plugin_arguments(['prepare_capture', 'capture', 'finish_capture'],
                              parser)
        if 'driver' in config.keys():
            _add_device_arguments('capture', parser)

    postprocess_parser = subparsers.add_parser(
        'postprocess',
        help="Postprocess scanned images.")
    postprocess_parser.add_argument(
        "path", type=unicode, help="Project path")
    postprocess_parser.add_argument(
        "--jobs", "-j", dest="jobs", type=int, default=None,
        metavar="<int>", help="Number of concurrent processes")
    postprocess_parser.set_defaults(subcommand=postprocess)
    # Add arguments from plugins
    for parser in (postprocess_parser, wizard_parser):
        _add_plugin_arguments(['process'], parser)

    output_parser = subparsers.add_parser(
        'output',
        help="Generate output files.")
    output_parser.add_argument(
        "path", type=unicode, help="Project path")
    output_parser.set_defaults(subcommand=output)
    # Add arguments from plugins
    for parser in (output_parser, wizard_parser):
        _add_plugin_arguments(['output'], parser)

    # Add custom subcommands from plugins
    if config["plugins"].get():
        pluginmanager.map(lambda x, y: x.plugin.add_command_parser(y),
                          subparsers)
    return rootparser


def set_config_from_args(config, args):
    for argkey, value in args.__dict__.iteritems():
        if value is None or argkey == 'subcommand' or argkey.startswith('_'):
            continue
        if '.' in argkey:
            section, key = argkey.split('.')
            config[section][key] = value
        else:
            config[argkey] = value


def main():
    # Set to ERROR so we can see errors during plugin loading.
    logging.basicConfig(loglevel=logging.ERROR)

    # Lazy-load configuration
    config = confit.LazyConfig('spreads', __name__)
    config.read()
    setup_plugin_config(config)

    parser = setup_parser(config)
    args = parser.parse_args()
    # Set configuration from parsed arguments
    set_config_from_args(config, args)

    loglevel = config['loglevel'].as_choice({
        'none':     logging.NOTSET,
        'info':     logging.INFO,
        'debug':    logging.DEBUG,
        'warning':  logging.WARNING,
        'error':    logging.ERROR,
        'critical': logging.CRITICAL,
    })

    # Set up logger
    logger = logging.getLogger()
    if logger.handlers:
        for handler in logger.handlers:
            logger.removeHandler(handler)
    handler = ColourStreamHandler()
    handler.setLevel(loglevel)
    handler.setFormatter(logging.Formatter("%(name)s: %(message)s"))
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

    args.subcommand(config)



import unittest

import pytest
import spreads.vendor.confit as confit
from mock import call, MagicMock as Mock, patch

import spreads.plugin as plugin
from spreads.util import DeviceException


class TestPlugin(unittest.TestCase):
    def setUp(self):
        pass

    def test_get_driver(self):
        assert "dummy" in plugin.get_driver("dummy").names()

    @patch('spreads.plugin.get_driver')
    def test_get_devices(self, get_driver):
        cfg = Mock()
        cfg.keys.return_value = ["driver"]
        driver = Mock()
        usb_mock = Mock()
        plugin.usb.core.find = Mock(return_value=[usb_mock])
        get_driver.return_value = driver
        plugin.get_driver = get_driver
        plugin.get_devices(cfg)
        assert call(cfg["device"], usb_mock) in driver.driver.call_args_list
        assert driver.driver.match.call_args_list == [call(usb_mock)]

    @patch('spreads.plugin.get_driver')
    def test_no_devices(self, get_driver):
        cfg = Mock()
        driver = Mock()
        driver.driver.match = Mock(return_value=False)
        usb_mock = Mock()
        plugin.usb.core.find = Mock(return_value=[usb_mock])
        get_driver.return_value = driver
        plugin.get_driver = get_driver
        with pytest.raises(DeviceException) as excinfo:
            plugin.get_devices(cfg)



# -*- coding: utf-8 -*-

import logging
import subprocess

import wand.image
from concurrent import futures

from spreads.plugin import HookPlugin, PluginOption
from spreads.util import find_in_path, MissingDependencyException

logger = logging.getLogger('spreadsplug.autorotate')

if not find_in_path('exiftool'):
    raise MissingDependencyException("Could not find executable `exiftool`"
                                     " in $PATH. Please install the"
                                     " appropriate package(s)!")


def _get_exif_orientation(fname):
    return int(subprocess.check_output(['exiftool', '-Orientation', '-n',
                                        unicode(fname)]).split(":")[1].strip())


def rotate_image(path, rotation):
    with wand.image.Image(filename=unicode(path)) as img:
        logger.debug("Rotating image \'{0}\' by {1} degrees"
                     .format(path, rotation))
        if img.height > img.width:
            logger.info("Image already rotated, skipping...")
        else:
            img.rotate(rotation)
            img.save(filename=unicode(path))
    # Update EXIF orientation
    if not _get_exif_orientation(path) == 1:
        logger.debug("Updating EXIF information for image '{0}'".format(path))
        subprocess.check_output(['exiftool', '-Orientation=1', '-n',
                                 '-overwrite_original', unicode(path)])


class AutoRotatePlugin(HookPlugin):
    __name__ = 'autorotate'

    @classmethod
    def configuration_template(cls):
        conf = {'rotate_odd': PluginOption(-90,
                                           "Rotation applied to odd pages"),
                'rotate_even': PluginOption(90,
                                            "Rotation applied to even pages"),
                }
        return conf

    def process(self, path):
        img_dir = path / 'raw'
        # Silence Wand logger
        (logging.getLogger("wand")
                .setLevel(logging.ERROR))
        logger.info("Rotating images in {0}".format(img_dir))
        with futures.ProcessPoolExecutor() as executor:
            for imgpath in sorted(img_dir.iterdir()):
                try:
                    exif_orientation = _get_exif_orientation(imgpath)
                    if exif_orientation == 8:
                        rotation = self.config['rotate_odd'].get(int)
                    elif exif_orientation == 6:
                        rotation = self.config['rotate_even'].get(int)
                    elif exif_orientation == 1:
                        # Already rotated, so we skip it
                        continue
                    else:
                        raise Exception("Invalid value for orientation: {0}"
                                        .format(exif_orientation))
                except Exception as exc:
                    logger.warn("Cannot determine rotation for image {0}!"
                                .format(imgpath))
                    logger.exception(exc)
                    continue
                logger.debug("Orientation for \"{0}\" is {1}"
                             .format(imgpath, rotation))
                executor.submit(rotate_image, imgpath, rotation=rotation)



import time
import unittest
from itertools import chain, repeat

import pytest
import spreads.vendor.confit as confit
from mock import patch, MagicMock as Mock

import spreads
import spreads.cli as cli
from spreads.util import DeviceException

spreads.util.find_in_path = Mock(return_value=True)


class TestCLI(unittest.TestCase):
    def setUp(self):
        self.workflow = Mock()
        self.workflow.devices = [Mock(), Mock()]
        self.workflow.devices[0].target_page = 'even'
        self.workflow.devices[1].target_page = 'odd'
        self.workflow.capture_start = time.time()
        self.workflow.config = confit.Configuration('test_cli')
        cli.Workflow = Mock(return_value=self.workflow)

    def test_capture(self):
        self.workflow.config['capture']['capture_keys'] = ["b", " "]
        cli.getch = Mock(side_effect=chain(repeat('b', 3), 'f'))
        cli.capture(self.workflow)
        assert cli.getch.call_count == 4
        assert self.workflow.prepare_capture.call_count == 1
        assert self.workflow.capture.call_count == 3
        assert self.workflow.finish_capture.call_count == 1
        #TODO: stats correct?

    def test_capture_nodevices(self):
        cli.getch = Mock(return_value=' ')
        self.workflow.devices = []
        with pytest.raises(DeviceException) as excinfo:
            cli.capture(self.workflow)

    def test_capture_no_target_page(self):
        self.workflow.devices[0].target_page = None
        cli.getch = Mock(return_value='c')
        with pytest.raises(DeviceException) as excinfo:
            cli.capture(self.workflow)

    def test_postprocess(self):
        self.workflow.path = '/tmp/foo'
        cli.postprocess(self.workflow)
        assert self.workflow.process.call_count == 1

    @patch('spreads.cli.capture')
    @patch('spreads.cli.postprocess')
    @patch('spreads.cli.output')
    def test_wizard(self, capture, postprocess, output):
        self.workflow.config['path'] = '/tmp/foo'
        cli.wizard(self.workflow.config)
        capture.assert_called_with(self.workflow.config)
        postprocess.assert_called_with(self.workflow.config)
        output.assert_called_with(self.workflow.config)


    def test_parser(self):
        cli.get_pluginmanager = Mock()
        # TODO: Test if plugin arguments are added
        parser = cli.setup_parser()

    @patch('os.path.exists')
    def test_main(self, exists):
        # TODO: Config dumped?
        # TODO: Config from args?
        # TODO: Loglevel set correctly?
        # TODO: Correct function executed?
        self.workflow.config["loglevel"] = "info"
        self.workflow.config["verbose"] = False
        self.workflow.config.dump = Mock()
        cli.confit.LazyConfig = Mock(return_value=self.workflow.config)
        cli.test_cmd = Mock()
        cli.setup_plugin_config = Mock()
        mock_parser = Mock()
        mock_args = Mock()
        mock_args.verbose = False
        mock_args.subcommand = Mock()
        mock_parser.parse_args = Mock(return_value=mock_args)
        cli.setup_parser = Mock(return_value=mock_parser)
        cli.set_config_from_args = Mock()
        exists.return_value = False
        cli.os.path.exists = exists
        cli.main()
        assert mock_args.subcommand.call_count == 1



# -*- coding: utf-8 -*-

# Copyright (c) 2013 Johannes Baiter. All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
spreads workflow object.
"""

from __future__ import division, unicode_literals

import logging
import time

import spreads.vendor.confit as confit
from concurrent.futures import ThreadPoolExecutor
from spreads.vendor.pathlib import Path

from spreads.plugin import get_pluginmanager, get_devices
from spreads.util import check_futures_exceptions, DeviceException


class Workflow(object):
    path = None
    step = None
    step_done = False
    capture_start = None
    _pages_shot = 0

    _devices = None
    _pluginmanager = None

    def __init__(self, path, config=None, step=None, step_done=None):
        self.logger = logging.getLogger('Workflow')
        self.step = step
        self.step_done = step_done
        if not isinstance(path, Path):
            path = Path(path)
        self.path = path
        if not self.path.exists():
            self.path.mkdir()
        # See if supplied `config` is already a valid Configuration object
        if isinstance(config, confit.Configuration):
            self.config = config
        else:
            self.config = self._load_config(config)

    @property
    def plugins(self):
        if self._pluginmanager is None:
            self._pluginmanager = get_pluginmanager(self.config)
        return [ext.obj for ext in self._pluginmanager]

    @property
    def devices(self):
        if self._devices is None:
            self._devices = get_devices(self.config)
        if not self._devices:
            raise DeviceException("Could not find any compatible devices!")
        return self._devices

    @property
    def images(self):
        # Get fresh image list if number of pages has changed
        raw_path = self.path / 'raw'
        if not raw_path.exists():
            return None
        return sorted(raw_path.iterdir())

    @property
    def out_files(self):
        out_path = self.path / 'out'
        if not out_path.exists():
            return None
        else:
            return sorted(out_path.iterdir())

    def _load_config(self, value):
        # Load default configuration
        config = confit.Configuration('spreads')
        cfg_file = self.path / 'config.yml'
        if value is None and cfg_file.exists():
            # Load workflow-specific configuration from file
            config.set(confit.ConfigSource({}, unicode(cfg_file)))
        elif value is not None:
            # Load configuration from supplied ConfigSource or dictionary
            config.set(value)
        return config

    def _run_hook(self, hook_name, *args):
        self.logger.debug("Running '{0}' hooks".format(hook_name))
        for plugin in self.plugins:
            getattr(plugin, hook_name)(*args)

    def _get_next_filename(self, target_page=None):
        """ Get next filename that a capture should be stored as.

        If the workflow is shooting with two devices, this will select a
        filename that matches the device's target page (odd/even).

        :param target_page: target page of file ('odd/even')
        :type target_page:  str/unicode/None if not applicable
        :return:            absolute path to next filename
                            (e.g. /tmp/proj/003.jpg)
        :rtype:             pathlib.Path
        """
        base_path = self.path / 'raw'
        if not base_path.exists():
            base_path.mkdir()

        try:
            last_num = int(self.images[-1].stem)
        except IndexError:
            last_num = -1

        if target_page is None:
            return base_path / "{03:0}".format(self._pages_shot)

        next_num = (last_num+2 if target_page == 'odd' else last_num+1)
        return base_path / "{0:03}".format(next_num)

    def prepare_capture(self):
        self.logger.info("Preparing capture.")
        self.step = 'capture'
        if any(dev.target_page is None for dev in self.devices):
            raise DeviceException(
                "Target page for at least one of the devicescould not be"
                "determined, please run 'spread configure' to configure your"
                "your devices.")
        with ThreadPoolExecutor(len(self.devices)) as executor:
            futures = []
            self.logger.debug("Preparing capture in devices")
            for dev in self.devices:
                futures.append(executor.submit(dev.prepare_capture, self.path))
        check_futures_exceptions(futures)

        flip_target = ('flip_target_pages' in self.config['device'].keys()
                       and self.config['device']['flip_target_pages'].get())
        if flip_target:
            (self.devices[0].target_page,
             self.devices[1].target_page) = (self.devices[1].target_page,
                                             self.devices[0].target_page)
        self._run_hook('prepare_capture', self.devices, self.path)

    def capture(self, retake=False):
        if self.capture_start is None:
            self.capture_start = time.time()
        self.logger.info("Triggering capture.")
        parallel_capture = ('parallel_capture' in self.config['device'].keys()
                            and self.config['device']['parallel_capture'].get()
                            )
        num_devices = 1 if not parallel_capture else 2

        if retake:
            # Remove last n images, where n == len(self.devices)
            map(lambda x: x.unlink(), self.images[-num_devices:])

        with ThreadPoolExecutor(num_devices) as executor:
            futures = []
            self.logger.debug("Sending capture command to devices")
            for dev in self.devices:
                img_path = self._get_next_filename(dev.target_page)
                futures.append(executor.submit(dev.capture, img_path))
        check_futures_exceptions(futures)
        self._run_hook('capture', self.devices, self.path)
        self._pages_shot += len(self.devices)

    def finish_capture(self):
        self.step_done = True
        self._run_hook('finish_capture', self.devices, self.path)

    def process(self):
        self.step = 'process'
        self.step_done = False
        self.logger.info("Starting postprocessing...")
        self._run_hook('process', self.path)
        self.logger.info("Done with postprocessing!")
        self.step_done = True

    def output(self):
        self.logger.info("Generating output files...")
        self.step = 'output'
        self.step_done = False
        out_path = self.path / 'out'
        if not out_path.exists():
            out_path.mkdir()
        self._run_hook('output', self.path)
        self.logger.info("Done generating output files!")
        self.step_done = True



# -*- coding: utf-8 -*-
import logging
import os
import re
import subprocess
import tempfile
import time
from fractions import Fraction
from itertools import chain

from spreads.vendor.pathlib import Path

from spreads.plugin import DevicePlugin, PluginOption, DeviceFeatures
from spreads.util import (DeviceException, find_in_path,
                          MissingDependencyException)

if not find_in_path('exiftool'):
    raise MissingDependencyException("Could not find executable `exiftool`"
                                     " in $PATH. Please install the"
                                     " appropriate package(s)!")


class CHDKPTPException(Exception):
    pass


class CHDKCameraDevice(DevicePlugin):
    """ Plugin for digital cameras running the CHDK firmware.

    """

    features = (DeviceFeatures.PREVIEW, DeviceFeatures.IS_CAMERA)

    target_page = None
    _cli_flags = None
    _chdk_buildnum = None
    _can_remote = False
    _zoom_steps = 0

    @classmethod
    def configuration_template(cls):
        conf = super(CHDKCameraDevice, cls).configuration_template()
        conf.update(
            {'sensitivity': PluginOption(80, "The ISO sensitivity value"),
             'shutter_speed': PluginOption(
                 u"1/25", "The shutter speed as a fraction"),
             'zoom_level': PluginOption(3, "The default zoom level"),
             'dpi': PluginOption(300, "The capturing resolution"),
             'shoot_raw': PluginOption(False, "Shoot in RAW format (DNG)"),
             'focus_distance': PluginOption('auto', "Set focus distance"),
             'chdkptp_path': PluginOption(
                 u"/usr/local/lib/chdkptp",
                 "Path to CHDKPTP binary/libraries"),
             })
        return conf

    @classmethod
    def match(cls, device):
        cfg = device.get_active_configuration()[(0, 0)]
        matches = (hex(cfg.bInterfaceClass) == "0x6"
                   and hex(cfg.bInterfaceSubClass) == "0x1")
        return matches

    def __init__(self, config, device):
        """ Set connection information and try to obtain target page.

        :param config:  spreads configuration
        :type config:   spreads.confit.ConfigView
        :param device:  USB device to use for the object
        :type device:   `usb.core.Device <http://github.com/walac/pyusb>`_

        """
        self.logger = logging.getLogger('ChdkCamera')
        self._chdkptp_path = Path(config["chdkptp_path"].get(unicode))
        self._sensitivity = config["sensitivity"].get(int)
        self._shutter_speed = float(Fraction(config["shutter_speed"]
                                    .get(unicode)))
        self._zoom_level = config['zoom_level'].get(int)
        self._dpi = config['dpi'].get(int)
        self._shoot_raw = config['shoot_raw'].get(bool)
        self._focus_distance = config['focus_distance'].get()

        self._cli_flags = []
        self._cli_flags.append("-c-d={0:03} -b={1:03}".format(
                               device.address, device.bus))
        self._cli_flags.append("-eset cli_verbose=2")
        self._chdk_buildnum = (self._execute_lua("get_buildinfo()",
                                                 get_result=True)
                               ["build_revision"])
        # PTP remote shooting is available starting from SVN r2927
        self._can_remote = self._chdk_buildnum >= 2927
        self._zoom_steps = self._execute_lua("get_zoom_steps()",
                                             get_result=True)
        try:
            self.target_page = self._get_target_page()
        except:
            self.target_page = None
        self.logger = logging.getLogger('ChdkCamera[{0}]'
                                        .format(self.target_page))

    def set_target_page(self, target_page):
        """ Set the device target page.

        :param target_page: The target page name
        :type target_page:  unicode in (u"odd", u"even")

        """
        tmp_handle = tempfile.mkstemp(text=True)
        os.write(tmp_handle[0], target_page.upper()+"\n")
        self._run("upload {0} \"OWN.TXT\"".format(tmp_handle[1]))
        self.target_page = target_page

    def prepare_capture(self, path):
        # Try to go into alt mode to prevent weird behaviour
        self._execute_lua("enter_alt()")
        # Try to put into record mode
        try:
            self._run("rec")
        except CHDKPTPException as e:
            self.logger.debug(e)
            self.logger.info("Camera already seems to be in recording mode")
        self._set_zoom(self._zoom_level)
        # Disable flash
        self._execute_lua("while(get_flash_mode()<2) do click(\"right\") end")
        # Disable ND filter
        self._execute_lua("set_nd_filter(2)")
        self._set_focus()

    def get_preview_image(self):
        fpath = tempfile.mkstemp()[1]
        cmd = "dumpframes -count=1 -nobm -nopal"
        self._run("{0} {1}".format(cmd, fpath))
        with open(fpath, 'rb') as fp:
            data = fp.read()
        os.remove(fpath)
        return data

    def capture(self, path, two_device_mode=False):
        # NOTE: To obtain the "real" Canon ISO value, we multiply the
        #       "market" value from the config by 0.65.
        #       See core/shooting.c#~l150 in the CHDK source for more details
        if self._can_remote:
            cmd = ("remoteshoot -tv={0} -sv={1} {2} {3}"
                   .format(self._shutter_speed, self._sensitivity*0.65,
                           "-dng" if self._shoot_raw else "", path))
        else:
            cmd = ("shoot -tv={0} -sv={1} -dng={2} -rm -dl {3}"
                   .format(self._shutter_speed, self._sensitivity*0.65,
                           int(self._shoot_raw), path))
        try:
            self._run(cmd)
        except CHDKPTPException as e:
            self.logger.warn("Capture command failed.")
            raise e

        extension = 'dng' if self._shoot_raw else 'jpg'
        local_path = "{0}.{1}".format(path, extension)
        # Set EXIF orientation
        self.logger.debug("Setting EXIF orientation on captured image")
        if self.target_page == 'odd':
            exif_orientation = 8  # 90°
        else:
            exif_orientation = 6  # -90°
        subprocess.check_output(
            ['exiftool', '-Orientation={0}'.format(exif_orientation), '-n',
             '-overwrite_original', local_path])

    def _run(self, *commands):
        cmd_args = list(chain((unicode(self._chdkptp_path / "chdkptp"),),
                              self._cli_flags,
                              ("-e{0}".format(cmd) for cmd in commands)))
        env = {'LUA_PATH': unicode(self._chdkptp_path / "lua/?.lua")}
        self.logger.debug("Calling chdkptp with arguments: {0}"
                          .format(cmd_args))
        output = (subprocess.check_output(cmd_args, env=env,
                                          stderr=subprocess.STDOUT)
                  .splitlines())
        self.logger.debug("Call returned:\n{0}".format(output))
        # Filter out connected message
        output = [x for x in output if not x.startswith('connected:')]
        # Check for possible CHDKPTP errors
        if any('ERROR' in x for x in output):
            raise CHDKPTPException("\n".join(output))
        return output

    def _execute_lua(self, script, wait=True, get_result=False, timeout=256):
        if get_result and not "return" in script:
            script = "return({0})".format(script)
        cmd = "luar" if wait else "lua"
        output = self._run("{0} {1}".format(cmd, script))
        if not get_result:
            return
        output = [x for x in output if x.find(":return:")][0]
        return self._parse_lua_output(output)

    def _parse_table(self, data):
        values = dict(re.findall(r'([\w_]+?)=(\d+|".*?"),', data[6:]))
        for k, v in values.iteritems():
            if v.startswith('"') and v.endswith('"'):
                values[k] = v.strip('"')  # String
            else:
                values[k] = int(v)  # Integer
        return values

    def _parse_lua_output(self, output):
        ret_val = re.match(r'^\d+:return:(.*)', output).group(1)
        if ret_val.startswith('table:'):
            return self._parse_table(ret_val)  # Table
        elif ret_val.startswith("'"):
            return ret_val.strip("'")  # String
        else:
            return int(ret_val)  # Integer

    def _get_target_page(self):
        tmp_handle = tempfile.mkstemp(text=True)
        try:
            self._run("download \"OWN.TXT\" {0}".format(tmp_handle[1]))
        except DeviceException:
            raise ValueError("Could not find OWN.TXT")
        with open(tmp_handle[1], 'r') as fp:
            target_page = fp.readline().strip().lower()
        os.remove(tmp_handle[1])
        if not target_page:
            raise ValueError("Could not read OWN.TXT")
        return target_page

    def _set_zoom(self, level):
        if level >= self._zoom_steps:
            raise Exception("Zoom level {0} exceeds the camera's range!"
                            " (max: {1})".format(level, self._zoom_steps-1))
        self._execute_lua("set_zoom({0})".format(level), wait=False)

    def _acquire_focus(self):
        """ Acquire auto focus and lock it. """
        self._execute_lua("enter_alt()")
        # Try to put into record mode
        try:
            self._run("rec")
        except CHDKPTPException as e:
            self.logger.debug(e)
            self.logger.info("Camera already seems to be in recording mode")
        self._set_zoom(self._zoom_level)
        self._execute_lua("set_aflock(0)")
        self._execute_lua("press('shoot_half')")
        time.sleep(0.8)
        self._execute_lua("release('shoot_half')")
        time.sleep(0.5)
        return self._execute_lua("get_focus()", get_result=True)

    def _set_focus(self):
        self._execute_lua("set_aflock(0)")
        if self._focus_distance == 'auto':
            return
        self._execute_lua("set_focus({0:.0f})".format(self._focus_distance))
        time.sleep(0.5)
        self._execute_lua("press('shoot_half')")
        time.sleep(0.25)
        self._execute_lua("release('shoot_half')")
        time.sleep(0.25)
        self._execute_lua("set_aflock(1)")


class CanonA2200CameraDevice(CHDKCameraDevice):
    """ Canon A2200 driver.

        Works around some quirks of that CHDK port.

    """
    def __init__(self, config, device):
        print "Instantiating device..."
        super(CanonA2200CameraDevice, self).__init__(config, device)
        if self.target_page is not None:
            self.logger = logging.getLogger(
                'CanonA2200CameraDevice[{0}]'.format(self.target_page))
        else:
            self.logger = logging.getLogger('CanonA2200CameraDevice')

    @classmethod
    def match(cls, device):
        matches = (hex(device.idVendor) == "0x4a9"
                   and hex(device.idProduct) == "0x322a")
        return matches

    def _set_zoom(self, level):
        """ Set zoom level.

            The A2200 currently has a bug, where setting the zoom level
            directly via set_zoom crashes the camera quite frequently, so
            we work around that by simulating button presses.

        :param level: The zoom level to be used
        :type level:  int

        """
        if level >= self._zoom_steps:
            raise DeviceException(
                "Zoom level {0} exceeds the camera's range!"
                " (max: {1})".format(level, self._zoom_steps-1))
        zoom = self._execute_lua("get_zoom()", get_result=True)
        if zoom < level:
            self._execute_lua("while(get_zoom()<{0}) do click(\"zoom_in\") end"
                              .format(level+1),
                              wait=True)
        elif zoom > level:
            self._execute_lua("while(get_zoom()>{0}) "
                              "do click(\"zoom_out\") end".format(level+1),
                              wait=True)



# -*- coding: utf-8 -*-

from __future__ import division, unicode_literals

import logging
import os
import subprocess

from spreads.plugin import HookPlugin
from spreads.util import MissingDependencyException, find_in_path

if not find_in_path('djvubind'):
    raise MissingDependencyException("Could not find executable `djvubind` in"
                                     " $PATH. Please install the appropriate"
                                     " package(s)!")

logger = logging.getLogger('spreadsplug.djvubind')


class DjvuBindPlugin(HookPlugin):
    __name__ = 'djvubind'

    def output(self, path):
        logger.info("Assembling DJVU.")
        img_dir = path / 'done'
        djvu_file = path / 'out' / "{0}.djvu".format(path.name)
        cmd = ["djvubind", unicode(img_dir)]
        if not img_dir.glob("*.html"):
            cmd.append("--no-ocr")
        logger.debug("Running " + " ".join(cmd))
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        os.rename("book.djvu", unicode(djvu_file))



import logging
import multiprocessing
import os
import re
import subprocess
import time
import xml.etree.cElementTree as ET

from spreads.plugin import HookPlugin, PluginOption
from spreads.util import find_in_path, MissingDependencyException
from spreads.vendor.pathlib import Path

if not find_in_path('tesseract'):
    raise MissingDependencyException("Could not find executable `tesseract`"
                                     " in $PATH. Please install the"
                                     " appropriate package(s)!")

logger = logging.getLogger('spreadsplug.tesseract')
try:
    AVAILABLE_LANGS = (subprocess.check_output(["tesseract", "--list-langfoo"],
                                               stderr=subprocess.STDOUT)
                       .split("\n")[1:-1])
except subprocess.CalledProcessError:
    AVAILABLE_LANGS = [x.stem for x in
                       Path('/usr/share/tesseract-ocr/tessdata')
                       .glob('*.traineddata')]


class TesseractPlugin(HookPlugin):
    __name__ = 'tesseract'

    @classmethod
    def configuration_template(cls):
        conf = {'language': PluginOption(value=AVAILABLE_LANGS,
                                         docstring="OCR language",
                                         selectable=True),
                }
        return conf

    def process(self, path):
        def _clean_processes(procs):
            for p in processes[:]:
                if p.poll() is not None:
                    processes.remove(p)

        ocr_lang = self.config['language'].get()
        logger.info("Performing OCR")
        logger.info("Language is \"{0}\"".format(ocr_lang))
        img_dir = path / 'done'

        processes = []
        max_procs = multiprocessing.cpu_count()
        FNULL = open(os.devnull, 'w')
        for img in img_dir.glob('*.tif'):
            # Wait until another process has finished
            while len(processes) >= max_procs:
                _clean_processes(processes)
                time.sleep(0.01)
            proc = subprocess.Popen(["tesseract", unicode(img),
                                     unicode(img_dir / img.stem), "-l",
                                     ocr_lang, "hocr"], stderr=FNULL,
                                     stdout=FNULL)
            processes.append(proc)
        # Wait for remaining processes to finish
        while processes:
            _clean_processes(processes)

        # NOTE: This modifies the hOCR files to make them compatible with
        #       pdfbeads.
        #       See the following bugreport for more information:
        #       http://rubyforge.org/tracker/index.php?func=detail&aid=29737&group_id=9752&atid=37737
        for fname in img_dir.glob('*.html'):
            with fname.open('r') as fp:
                new_content = re.sub(r'(<span[^>]*>(<strong>)? +(<\/strong>)?'
                                     r'<\/span>*)(<span[^>]*>(<strong>)? '
                                     r'+(<\/strong>)?<\/span> *)',
                                     r'\g<1>', fp.read())
            with fname.open('w') as fp:
                fp.write(new_content)

    def output(self, path):
        outfile = path / 'out' / "{0}.hocr".format(path.name)
        inpath = path / 'done'
        out_root = ET.Element('html')
        ET.SubElement(out_root, 'head')
        body = ET.SubElement(out_root, 'body')

        for idx, page in enumerate(sorted(inpath.glob('*.html'))):
            # NOTE: Fixe some things from hOCR output so we can parse it...
            with page.open('r+') as fp:
                content = re.sub(r'<em><\/em>', '', fp.read())
                content = re.sub(r'<strong><\/strong>', '', content)
            page = ET.fromstring(content.encode('utf-8'))
            page_elem = page.find("xhtml:body/xhtml:div[@class='ocr_page']",
                                  dict(xhtml="http://www.w3.org/1999/xhtml"))
            # Correct page_number
            page_elem.set('id', 'page_{0}'.format(idx))

            # And tack it onto the output file
            if page_elem:
                body.append(page_elem)
        with outfile.open('w') as fp:
            # Strip those annoying namespace tags...
            out_string = re.sub(r"(<\/*)html:", "\g<1>",
                                ET.tostring(out_root))
            fp.write(unicode(out_string))



import logging
import time

from concurrent.futures import ThreadPoolExecutor
from PySide import QtCore, QtGui

import spreads.workflow as workflow
from spreads.plugin import get_pluginmanager

import gui_rc

logger = logging.getLogger('spreadsplug.gui')


class LogBoxHandler(logging.Handler):
    # NOTE: This is neccessary, because a signal has to be an attribute of
    #       a QObject instance. Multiple inheritance does not work here,
    #       as both logging.Handler and QObject have an "emit" method
    class DummyQObject(QtCore.QObject):
        sig = QtCore.Signal(unicode)

    def __init__(self, textedit):
        logging.Handler.__init__(self)
        self._qobj = self.DummyQObject()
        self.sig = self._qobj.sig
        self.textedit = textedit

    def emit(self, record):
        # Play a warning sound if something has gone wrong
        try:
            self.sig.emit(self.format(record))
        except:
            self.handleError(record)

        if record.levelname in ('ERROR', 'CRITICAL', 'WARNING'):
            QtGui.QApplication.beep()


class LogBoxFormatter(logging.Formatter):
    LEVELS = {
        'ERROR': "<strong><font color=\"red\">{0}</font></strong>",
        'CRITICAL': "<strong><font color=\"orange\">{0}</font></strong>",
        'WARNING': "<strong><font color=\"orange\">{0}</font></strong>",
        'INFO': "{0}",
        'DEBUG': "{0}",
    }

    def format(self, record):
        levelname = record.levelname
        return self.LEVELS[levelname].format(record.msg)


class SpreadsWizard(QtGui.QWizard):
    def __init__(self, config, parent=None):
        super(SpreadsWizard, self).__init__(parent)

        self.config = config

        self.addPage(IntroPage())
        self.addPage(CapturePage())
        self.addPage(PostprocessPage())
        self.addPage(OutputPage())

        self.setPixmap(QtGui.QWizard.WatermarkPixmap,
                       QtGui.QPixmap(':/pixmaps/monk.png'))

        button_layout = []
        button_layout.append(QtGui.QWizard.BackButton)
        button_layout.append(QtGui.QWizard.Stretch)
        button_layout.append(QtGui.QWizard.CancelButton)
        button_layout.append(QtGui.QWizard.Stretch)
        button_layout.append(QtGui.QWizard.NextButton)
        button_layout.append(QtGui.QWizard.FinishButton)
        self.setButtonLayout(button_layout)

        self.setWindowTitle("Spreads Wizard")
        self.setFixedWidth(700)


class IntroPage(QtGui.QWizardPage):
    def initializePage(self):
        wizard = self.wizard()
        self.pluginmanager = get_pluginmanager(wizard.config)
        wizard.active_plugins = self.pluginmanager.names()
        self.setTitle("Welcome!")

        intro_label = QtGui.QLabel(
            "This wizard will guide you through the digitization workflow. "
        )
        intro_label.setWordWrap(True)

        dirpick_layout = QtGui.QHBoxLayout()
        self.line_edit = QtGui.QLineEdit()
        self.line_edit.textChanged.connect(self.completeChanged)
        browse_btn = QtGui.QPushButton("Browse")
        dirpick_layout.addWidget(self.line_edit)
        dirpick_layout.addWidget(browse_btn)
        browse_btn.clicked.connect(self.show_filepicker)

        self.stack_widget = QtGui.QStackedWidget()
        page_combobox = QtGui.QComboBox()
        page_combobox.activated.connect(self.stack_widget.setCurrentIndex)
        #QtCore.QObject.connect(page_combobox, SIGNAL("activated(int)"),
        #        self.stack_widget, SLOT("setCurrentIndex(int)"))

        # Add configuration widgets from plugins
        self.plugin_widgets = {}
        for ext in self.pluginmanager:
            tmpl = ext.plugin.configuration_template()
            if not tmpl:
                continue
            page = QtGui.QGroupBox()
            layout = QtGui.QFormLayout()
            widgets = self._get_plugin_config_widgets(tmpl)
            self.plugin_widgets[ext.name] = widgets
            for label, widget in widgets.values():
                # We don't need a label for QCheckBoxes
                if isinstance(widget, QtGui.QCheckBox):
                    layout.addRow(widget)
                else:
                    layout.addRow(label, widget)
            page.setLayout(layout)
            self.stack_widget.addWidget(page)
            page_combobox.addItem(ext.name.title())

        main_layout = QtGui.QVBoxLayout()
        main_layout.addWidget(intro_label)
        main_layout.addSpacing(30)
        main_layout.addWidget(
            QtGui.QLabel("Please select a project directory.")
        )
        main_layout.addLayout(dirpick_layout)
        main_layout.addSpacing(30)
        main_layout.addWidget(page_combobox)
        main_layout.addWidget(self.stack_widget)
        main_layout.setSizeConstraint(QtGui.QLayout.SetNoConstraint)

        self.setLayout(main_layout)
        self.adjustSize()

    def _get_plugin_config_widgets(self, plugin_tmpl):
        widgets = {}
        for key, option in plugin_tmpl.items():
            label = (option.docstring
                     if not option.docstring is None else key.title())
            # Do we need a dropdown?
            if (option.selectable and
                    any(isinstance(option.value, x) for x in (list, tuple))):
                widget = QtGui.QComboBox()
                for each in option.value:
                    widget.addItem(each)
                widget.setCurrentIndex(0)
            # Do we need a checkbox?
            elif isinstance(option.value, bool):
                widget = QtGui.QCheckBox(label)
                widget.setCheckState(QtCore.Qt.Checked
                                     if option.value
                                     else QtCore.Qt.Unchecked)
            elif any(isinstance(option.value, x) for x in (list, tuple)):
                # NOTE: We skip options with sequences for a value for now,
                #       as I'm not sure yet how this is best displayed to
                #       the user...
                # TODO: Find a way to display this nicely to the user
                continue
            # Seems like we need another value...
            elif isinstance(option.value, int):
                widget = QtGui.QSpinBox()
                widget.setMaximum(1024)
                widget.setMinimum(-1024)
                widget.setValue(option.value)
            elif isinstance(option.value, float):
                widget = QtGui.QDoubleSpinBox()
                widget.setMaximum(1024.0)
                widget.setMinimum(-1024.0)
                widget.setValue(option.value)
            else:
                widget = QtGui.QLineEdit()
                widget.setText(option.value)
            widgets[key] = (label, widget)
        return widgets

    def show_filepicker(self):
        dialog = QtGui.QFileDialog()
        dialog.setFileMode(QtGui.QFileDialog.Directory)
        dialog.setOption(QtGui.QFileDialog.ShowDirsOnly, True)
        dialog.exec_()
        self.line_edit.setText(dialog.selectedFiles()[0])

    def isComplete(self):
        return bool(self.line_edit.text())

    def validatePage(self):
        wizard = self.wizard()
        project_path = self.line_edit.text()
        if not project_path:
            msg_box = QtGui.QMessageBox()
            msg_box.setText("Please select a project directory.")
            msg_box.setIcon(QtGui.QMessageBox.Critical)
            msg_box.exec_()
            return False

        self._update_config_from_plugin_widgets()
        wizard.workflow = workflow.Workflow(path=project_path,
                                            config=wizard.config)
        wizard.workflow.prepare_capture()
        return True

    def _update_config_from_plugin_widgets(self):
        config = self.wizard().config
        for ext in self.pluginmanager:
            if not ext.name in self.plugin_widgets:
                continue
            logger.debug("Updating config from widgets for plugin {0}"
                         .format(ext.name))
            for key in self.plugin_widgets[ext.name]:
                logger.debug("Trying to set key \"{0}\"".format(key))
                label, widget = self.plugin_widgets[ext.name][key]
                if isinstance(widget, QtGui.QComboBox):
                    idx = widget.currentIndex()
                    values = ext.plugin.configuration_template()[key].value
                    logger.debug("Setting to \"{0}\"".format(values[idx]))
                    config[ext.name][key] = values[idx]
                elif isinstance(widget, QtGui.QCheckBox):
                    checked = widget.isChecked()
                    logger.debug("Setting to \"{0}\"".format(checked))
                    config[ext.name][key] = checked
                elif any(isinstance(widget, x)
                         for x in (QtGui.QSpinBox, QtGui.QDoubleSpinBox)):
                    config[ext.name][key] = widget.value()
                else:
                    content = config[ext.name][key] = widget.text()
                    logger.debug("Setting to \"{0}\"".format(content))
                    config[ext.name][key] = unicode(content)


class CapturePage(QtGui.QWizardPage):
    def initializePage(self):
        self.setTitle("Capturing from devices")
        self.start_time = None
        self.shot_count = None

        # TODO: Add horizontally scrolling box with thumbnails of all
        #       previously shot images
        # TODO: Display last two shot images next to each other
        # TODO: Add button to retake the last capture

        layout = QtGui.QVBoxLayout(self)
        self.status = QtGui.QLabel("Press a capture key (default: Space, B)"
                                   " to begin capturing.")

        self.capture_btn = QtGui.QPushButton("Capture")
        self.capture_btn.clicked.connect(self.doCapture)
        self.capture_btn.setFocus()

        control_layout = QtGui.QHBoxLayout()
        self.control_odd = QtGui.QLabel()
        self.control_even = QtGui.QLabel()
        control_layout.addWidget(self.control_odd)
        control_layout.addWidget(self.control_even)

        self.retake_btn = QtGui.QPushButton("Retake")
        self.retake_btn.clicked.connect(self.retakeCapture)
        self.retake_btn = QtGui.QPushButton("Retake")
        self.retake_btn.clicked.connect(self.retakeCapture)

        self.logbox = QtGui.QTextEdit()
        self.log_handler = LogBoxHandler(self.logbox)
        self.log_handler.setLevel(logging.WARNING)
        self.log_handler.setFormatter(LogBoxFormatter())
        logging.getLogger().addHandler(self.log_handler)
        self.log_handler.sig.connect(self.logbox.append)

        layout.addWidget(self.status)
        layout.addStretch(1)
        layout.addLayout(control_layout)
        layout.addWidget(self.capture_btn)
        layout.addWidget(self.retake_btn)
        layout.addWidget(self.logbox)
        self.setLayout(layout)

    def validatePage(self):
        self.wizard().workflow.finish_capture()
        return True

    def keyPressEvent(self, event):
        if event.key() in (QtCore.Qt.Key_B, QtCore.Qt.Key_Space):
            self.doCapture()

    def updateControl(self):
        images = self.wizard().workflow.images
        self.control_odd.setPixmap(
            QtGui.QPixmap.fromImage(QtGui.QImage(unicode(images[-2]))
                                    .scaledToWidth(250)))
        self.control_even.setPixmap(
            QtGui.QPixmap.fromImage(QtGui.QImage(unicode(images[-1]))
                                    .scaledToWidth(250)))

    def retakeCapture(self):
        self.retake_btn.setEnabled(False)
        self.doCapture(retake=True)
        self.retake_btn.setEnabled(True)

    def doCapture(self, retake=False):
        self.capture_btn.setEnabled(False)
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(self.wizard().workflow.capture, retake)
            while future.running():
                QtGui.qApp.processEvents()
                time.sleep(0.001)
        self.capture_btn.setEnabled(True)
        self.capture_btn.setFocus()
        pages_shot = self.wizard().workflow._pages_shot
        start_time = self.wizard().workflow.capture_start
        self.status.setText(
            "Shot {0} pages in {1:.0f} minutes ({2:.0f} pages/hour)"
            .format(pages_shot, (time.time() - start_time) / 60,
                    ((3600 / (time.time() - start_time)) * pages_shot)))
        self.updateControl()


class PostprocessPage(QtGui.QWizardPage):
    def initializePage(self):
        self.setTitle("Postprocessing")

        self.done = False

        self.progressbar = QtGui.QProgressBar(self)
        self.progressbar.setRange(0, 0)
        self.progressbar.setAlignment(QtCore.Qt.AlignCenter)
        self.logbox = QtGui.QTextEdit()
        self.log_handler = LogBoxHandler(self.logbox)
        self.log_handler.setLevel(logging.INFO)
        self.log_handler.setFormatter(LogBoxFormatter())
        logging.getLogger().addHandler(self.log_handler)
        self.log_handler.sig.connect(self.logbox.append)

        layout = QtGui.QVBoxLayout()
        layout.addWidget(self.progressbar)
        layout.addWidget(self.logbox)
        self.setLayout(layout)
        self.logbox.clear()
        QtCore.QTimer.singleShot(0, self.doPostprocess)

    def doPostprocess(self):
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(self.wizard().workflow.process)
            while not future.done():
                QtGui.qApp.processEvents()
                self.progressbar.setValue(0)
                time.sleep(0.01)
            if future.exception():
                raise future.exception()
        self.progressbar.hide()
        self.done = True
        self.completeChanged.emit()

    def isComplete(self):
        return self.done

    def validatePage(self):
        logging.getLogger().removeHandler(self.log_handler)
        return True


class OutputPage(QtGui.QWizardPage):
    def initializePage(self):
        self.setTitle("Generating output files")
        self.setFinalPage(True)

        self.done = False

        self.progressbar = QtGui.QProgressBar(self)
        self.progressbar.setRange(0, 0)
        self.progressbar.setAlignment(QtCore.Qt.AlignCenter)
        self.logbox = QtGui.QTextEdit()
        self.log_handler = LogBoxHandler(self.logbox)
        self.log_handler.setLevel(logging.INFO)
        self.log_handler.setFormatter(LogBoxFormatter())
        logging.getLogger().addHandler(self.log_handler)
        self.log_handler.sig.connect(self.logbox.append)

        layout = QtGui.QVBoxLayout()
        layout.addWidget(self.progressbar)
        layout.addWidget(self.logbox)
        self.setLayout(layout)
        self.logbox.clear()
        QtCore.QTimer.singleShot(0, self.doGenerateOutput)

    def doGenerateOutput(self):
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(self.wizard().workflow.output)
            while not future.done():
                QtGui.qApp.processEvents()
                self.progressbar.setValue(0)
                time.sleep(0.001)
        self.progressbar.hide()
        self.done = True
        self.completeChanged.emit()

    def isComplete(self):
        return self.done

    def validatePage(self):
        logging.getLogger().removeHandler(self.log_handler)
        return True



# -*- coding: utf-8 -*-

# Copyright (c) 2013 Johannes Baiter. All rights reserved.
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import division, unicode_literals

import abc
import itertools
import logging
import os

import usb
import stevedore
from stevedore.driver import DriverManager
from stevedore.named import NamedExtensionManager

from spreads.util import abstractclassmethod, DeviceException

logger = logging.getLogger("spreads.plugin")


class PluginOption(object):
    """ A configuration option.

    :attr value:      The default value for the option or a list of available
                      options if :attr selectable: is True
    :type value:      object (or list/tuple when :attr selectable: is True)
    :attr docstring:  A string explaining the configuration option
    :type docstring:  unicode
    :attr selectable: Make the PluginOption a selectable, i.e. value contains
                      a list or tuple of acceptable values for this option,
                      with the first member being the default selection.
    :type selectable: bool
    """

    def __init__(self, value, docstring=None, selectable=False):
        self.value = value
        self.docstring = docstring
        self.selectable = selectable


class SpreadsNamedExtensionManager(NamedExtensionManager):
    """ Custom extension manager for spreads.

    stevedore's NamedExtensionmanger does not give us the Exception that caused
    a plugin to fail at initialization. This derived class throws the original
    exception instead of logging it.
    """
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = (super(SpreadsNamedExtensionManager, cls)
                             .__new__(cls, *args, **kwargs))
        return cls._instance

    def _load_plugins(self, invoke_on_load, invoke_args, invoke_kwds):
        extensions = []
        for ep in self._find_entry_points(self.namespace):
            stevedore.LOG.debug('found extension %r', ep)
            ext = self._load_one_plugin(ep,
                                        invoke_on_load,
                                        invoke_args,
                                        invoke_kwds,
                                        )
            if ext:
                extensions.append(ext)
        return extensions


class SpreadsPlugin(object):
    """ Plugin base class.

    """
    @classmethod
    def configuration_template(cls):
        """ Allows a plugin to define its configuration keys.

        The returned dictionary has to be flat (i.e. no nested dicts)
        and contain a PluginOption object for each key.

        Example::

          {
           'a_setting': PluginOption(value='default_value'),
           'another_setting': PluginOption(value=[1, 2, 3],
                                           docstring="A list of things"),
           # In this case, 'full-fat' would be the default value
           'milk': PluginOption(value=('full-fat', 'skim'),
                                docstring="Type of milk",
                                selectable=True),
          }

        :return: dict with unicode: PluginOption(value, docstring, selection)
        """
        pass

    def __init__(self, config):
        """ Initialize the plugin.

        :param config: The global configuration object, by default only the
                       section with plugin-specific values gets stored in
                       the `config` attribute, if the plugin has a `__name__`
                       attribute.
        :type config: confit.ConfigView

        """
        if hasattr(self, '__name__'):
            self.config = config[self.__name__]
        else:
            self.config = config


class DeviceFeatures(object):
    #: Device can grab a preview picture
    PREVIEW = 1

    #: Device class allows the operation of two devices simultaneously
    #: (mainly to be used by cameras, where each device is responsible for
    #: capturing a single page.
    IS_CAMERA = 2


class DevicePlugin(SpreadsPlugin):
    """ Base class for devices.

        Subclass to implement support for different devices.

    """
    __metaclass__ = abc.ABCMeta

    #: Tuple of :py:class:`DeviceFeatures` constants that designate the
    #: features the device offers.
    features = ()

    @classmethod
    def configuration_template(cls):
        if DeviceFeatures.IS_CAMERA in cls.features:
            return {
                "parallel_capture": PluginOption(
                    value=True,
                    docstring="Trigger capture on multiple devices at once.",
                    selectable=False),
                "flip_target_pages": PluginOption(
                    value=False, docstring="Temporarily switch target pages"
                                           "(useful for e.g. East-Asian books")
            }

    @abstractclassmethod
    def match(cls, usbdevice):
        """ Match device against USB device information.

        :param device:      The USB device to match against
        :type vendor_id:    `usb.core.Device <http://github.com/walac/pyusb>`_
        :return:            True if the :param usbdevice: matches the
                            implemented category

        """
        raise NotImplementedError

    def __init__(self, config, device):
        """ Set connection information and other properties.

        :param config:  spreads configuration
        :type config:   spreads.confit.ConfigView
        :param device:  USB device to use for the object
        :type device:   `usb.core.Device <http://github.com/walac/pyusb>`_

        """
        super(DevicePlugin, self).__init__(config['device'])
        self._device = device

    def set_target_page(self, target_page):
        """ Set the device target page, if applicable.

        :param target_page: The target page
        :type target_page:  unicode in (u"odd", u"even")

        """
        raise NotImplementedError

    @abc.abstractmethod
    def prepare_capture(self, path):
        """ Prepare device for scanning.

        What this means exactly is up to the implementation and the type,
        of device, usually it involves things like switching into record
        mode, path and applying all relevant settings.

        :param path:    Project base path
        :type path:     pathlib.Path

        """
        raise NotImplementedError

    @abc.abstractmethod
    def capture(self, path):
        """ Capture a single image with the device.

        :param path:    Path for the image
        :type path:     pathlib.Path

        """
        raise NotImplementedError


class HookPlugin(SpreadsPlugin):
    """ Add functionality to any of spreads' commands by implementing one or
        more of the available hooks.

    """
    @classmethod
    def add_command_parser(cls, rootparser):
        """ Allows a plugin to register a new command with the command-line
            parser. The subparser that is added to :param rootparser: should
            set the class' ``__call__`` method as the ``func`` (via
            ``set_defaults``) that is executed when the subcommand is specified
            on the CLI.

        :param rootparser: The root parser that this plugin should add a
                           subparser to.
        :type rootparser:  argparse.ArgumentParser

        """
        pass

    def prepare_capture(self, devices, path):
        """ Perform some action before capturing begins.

        :param devices:     The devices used for capturing
        :type devices:      list(DevicePlugin)
        :param path:        Project path
        :type path:         pathlib.Path

        """
        pass

    def capture(self, devices, path):
        """ Perform some action after each successful capture.

        :param devices:     The devices used for capturing
        :type devices:      list(DevicePlugin)
        :param path:        Project path
        :type path:         pathlib.Path

        """
        pass

    def finish_capture(self, devices, path):
        """ Perform some action after capturing has finished.

        :param devices:     The devices used for capturing
        :type devices:      list(DevicePlugin)
        :param path:        Project path
        :type path:         pathlib.Path

        """
        pass

    def process(self, path):
        """ Perform one or more actions that either modify the captured images
            or generate a different output.

        .. note:
            This method is intended to operate on the *done* subdfolder of
            the project directory. At the beginning of postprocessing, it
            will contain copies of the images in *raw*. This is to ensure that
            a copy of the original, scanned images will always be available
            for archival purposes.

        :param path:        Project path
        :type path:         pathlib.Path

        """
        pass

    def output(self, path):
        """ Assemble an output file from the postprocessed images.

        .. note:
            This method is intended to take its input files from the *done*
            subfolder of the project path and store its output in the
            *out* subfolder.

        :param path:        Project path
        :type path:         pathlib.Path

        """
        pass


def get_pluginmanager(config):
    logger.debug("Creating plugin manager")
    pluginmanager = SpreadsNamedExtensionManager(
        namespace='spreadsplug.hooks',
        names=config['plugins'].as_str_seq(),
        invoke_on_load=True,
        invoke_args=[config],
        name_order=True)
    return pluginmanager


def get_driver(driver_name):
    return DriverManager(namespace="spreadsplug.devices",
                         name=driver_name)


def get_devices(config):
    """ Initialize configured devices.
    """
    if not 'driver' in config.keys():
        raise DeviceException(
            "No driver has been configured\n"
            "Please run `spread configure` to select a driver.")
    driver = get_driver(config["driver"].get())
    driver_class = driver.driver
    logger.debug("Finding devices for driver \"{0}\"".format(driver))
    usb_devices = filter(lambda dev: driver_class.match(dev),
                         usb.core.find(find_all=True))
    devices = [driver_class(config['device'], dev) for dev in usb_devices]
    if not devices:
        raise DeviceException("Could not find any compatible devices!")
    return devices


def setup_plugin_config(config):
    pluginmanager = get_pluginmanager(config)
    if "driver" in config.keys():
        driver = get_driver(config["driver"].get(unicode))
    else:
        driver = ()
    for ext in itertools.chain(pluginmanager, driver):
        logger.debug("Obtaining configuration template for plugin \"{0}\""
                     .format(ext.name))
        tmpl = ext.plugin.configuration_template()
        if ext in driver.extensions:
            section = 'device'
        else:
            section = ext.name
        if not tmpl:
            continue
        # Add default values
        for key, option in tmpl.items():
            # Check if we already have a configuration entry for this setting
            if key in config[section].keys():
                continue
            logging.info("Adding setting {0} from  plugin {1}"
                         .format(key, ext.name))
            if option.selectable:
                config[section][key] = option.value[0]
            else:
                config[section][key] = option.value


def get_relevant_extensions(plugin_manager, hooks):
    """ Find all extensions that implement certain hooks.

    :param hooks:   A list of hook method names
    :type hooks:    list(unicode)
    :return:        A generator that yields relevant extensions
    :rtype:         generator(Extension)

    """
    # NOTE: This one is wicked... The goal is to find all extensions that
    #       implement one of the specified hooks.
    #       To do so, we compare the code objects for the appropriate
    #       hook method with the same method in the HookPlugin base class.
    #       If the two are not the same, we can (somewhat) safely assume
    #       that the extension implements this hook and is thus relevant
    #       to us.
    #       Yes, this is not ideal and is due to our somewhat sloppy
    #       plugin interface. That's why...
    # TODO: Refactor plugin interface to make this less painful
    for ext in plugin_manager:
        relevant = any(
            getattr(ext.plugin, hook).func_code is not
            getattr(HookPlugin, hook).func_code
            for hook in hooks
        )
        if relevant:
            yield ext



from __future__ import division, unicode_literals

import unittest

import pytest
import spreads.vendor.confit as confit
from mock import call, MagicMock as Mock, patch
from spreads.vendor.pathlib import Path

import spreads.util as util
import spreads.workflow as workflow

util.find_in_path = Mock(return_value=True)


class TestWorkflow(unittest.TestCase):
    def setUp(self):
        self.plugins = [Mock() for x in xrange(3)]
        self.devices = [Mock() for x in xrange(2)]
        self.devices[0].target_page = 'odd'
        self.devices[1].target_page = 'even'
        workflow.get_pluginmanager = Mock(return_value=self.plugins)
        workflow.get_devices = Mock(return_value=self.devices)
        config = confit.Configuration('test_workflow')
        self.mock_path = Mock(wraps=Path('/tmp/test_workflow'), spec=Path)
        self.workflow = workflow.Workflow(config=config,
                                          path=self.mock_path)

    def test_get_plugins(self):
        foo = self.workflow.plugins
        bar = self.workflow.plugins
        assert workflow.get_pluginmanager.call_count == 1
        assert foo == bar == [x.obj for x in self.plugins]

    def test_get_devices(self):
        foo = self.workflow.devices
        bar = self.workflow.devices
        assert workflow.get_devices.call_count == 1
        assert foo == bar == self.devices

    def test_get_devices_no_device(self):
        workflow.get_devices = Mock(return_value=[])
        with pytest.raises(util.DeviceException) as excinfo:
            self.workflow.devices

    @patch('spreads.plugin.os.mkdir')
    def test_get_next_filename(self, mkdir):
        fname = self.workflow._get_next_filename(target_page='odd')
        assert fname == self.mock_path/'raw'/'001.jpg'
        fname = self.workflow._get_next_filename(target_page='even')
        assert fname == self.mock_path/'raw'/'000.jpg'
        self.workflow._pages_shot += 2
        fname = self.workflow._get_next_filename(target_page='odd')
        assert fname == self.mock_path/'raw'/'003.jpg'
        fname = self.workflow._get_next_filename(target_page='even')
        assert fname == self.mock_path/'raw'/'002.jpg'

    def test_prepare_capture(self):
        self.workflow.prepare_capture()
        for dev in self.devices:
            assert dev.prepare_capture.call_count == 1
            assert dev.prepare_capture.called_with(self.mock_path)
        for plug in self.plugins:
            assert plug.obj.prepare_capture.call_count == 1
            assert plug.obj.prepare_capture.call_args_list == (
                [call(self.devices, self.mock_path)])

    def test_capture(self):
        self.workflow.config['device']['parallel_capture'] = True
        self.workflow.config['device']['flip_target_pages'] = False
        self.workflow.capture()
        self.devices[0].capture.assert_called_with(self.mock_path/'raw'/'001.jpg')
        self.devices[1].capture.assert_called_with(self.mock_path/'raw'/'000.jpg')
        for plug in self.plugins:
            assert plug.obj.capture.call_count == 1
            assert (plug.obj.capture.call_args_list ==
                    [call(self.devices, self.mock_path)])

    def test_capture_noparallel(self):
        self.workflow.config['device']['parallel_capture'] = False
        self.workflow.config['device']['flip_target_pages'] = False
        self.workflow.capture()
        # TODO: Find a way to verify that the cameras were indeed triggered
        #       in sequence and not in parallel
        for dev in self.devices:
            assert dev.capture.call_count == 1

    def test_capture_flip_target_pages(self):
        self.workflow.config['device']['parallel_capture'] = False
        self.workflow.config['device']['flip_target_pages'] = True
        self.workflow.capture()
        self.devices[0].capture.assert_called_with(self.mock_path/'raw'/'000.jpg')
        self.devices[1].capture.assert_called_with(self.mock_path/'raw'/'001.jpg')

    def test_finish_capture(self):
        self.workflow.finish_capture()
        for plug in self.plugins:
            assert plug.obj.finish_capture.call_count == 1
            assert plug.obj.finish_capture.call_args_list == (
                [call(self.devices, self.mock_path)])

    def test_process(self):
        self.workflow.process()
        for plug in self.plugins:
            assert plug.obj.process.call_count == 1
            assert (plug.obj.process.call_args_list ==
                    [call(self.mock_path)])

    def test_output(self):
        (self.mock_path/'out').exists.return_value = False
        self.workflow.output()
        for plug in self.plugins:
            assert plug.obj.output.call_count == 1
            assert (plug.obj.output.call_args_list ==
                    [call(self.mock_path)])
        assert (self.mock_path/'out').mkdir.call_count == 1



# -*- coding: utf-8 -*-

from __future__ import division, unicode_literals

import logging
import math
import multiprocessing
import re
import shutil
import subprocess
import tempfile
from xml.etree.cElementTree import ElementTree as ET

from spreads.vendor.pathlib import Path

from spreads.plugin import HookPlugin, PluginOption
from spreads.util import find_in_path, MissingDependencyException

if not find_in_path('scantailor-cli'):
    raise MissingDependencyException("Could not find executable"
                                     " `scantailor-cli` in $PATH. Please"
                                     " install the appropriate package(s)!")

logger = logging.getLogger('spreadsplug.scantailor')


class ScanTailorPlugin(HookPlugin):
    __name__ = 'scantailor'

    _enhanced = bool(re.match(r".*<images\|directory\|->.*",
                              subprocess.check_output('scantailor-cli')
                              .splitlines()[7]))

    @classmethod
    def configuration_template(cls):
        conf = {'autopilot': PluginOption(value=False,
                                          docstring="Skip manual correction"),
                'rotate': PluginOption(value=False, docstring="Rotate pages"),
                'split_pages': PluginOption(value=True,
                                            docstring="Split pages"),
                'deskew': PluginOption(value=True, docstring="Deskew pages"),
                'content': PluginOption(value=True,
                                        docstring="Detect page content"),
                'auto_margins': PluginOption(value=True,
                                             docstring="Automatically detect"
                                                       " margins"),
                'detection': PluginOption(value=('content', 'page'),
                                          docstring="Content detection mode",
                                          selectable=True),
                'margins': PluginOption([2.5, 2.5, 2.5, 2.5])
                }
        return conf

    def _generate_configuration(self, projectfile, img_dir, out_dir):
        if not out_dir.exists():
            out_dir.mkdir()
        logger.info("Generating ScanTailor configuration")
        filterconf = [self.config[x].get(bool)
                      for x in ('rotate', 'split_pages', 'deskew', 'content',
                                'auto_margins')]
        start_filter = filterconf.index(True)+1
        end_filter = len(filterconf) - list(reversed(filterconf)).index(True)+1
        marginconf = self.config['margins'].as_str_seq()
        generation_cmd = ['scantailor-cli',
                          '--start-filter={0}'.format(start_filter),
                          '--end-filter={0}'.format(end_filter),
                          '--layout=1.5',
                          '-o={0}'.format(projectfile)]
        page_detection = self.config['detection'].get() == 'page'
        if self._enhanced and page_detection:
            generation_cmd.extend([
                '--enable-page-detection',
                '--disable-content-detection',
                '--enable-fine-tuning'
            ])
        else:
            generation_cmd.extend([
                '--margins-top={0}'.format(marginconf[0]),
                '--margins-right={0}'.format(marginconf[1]),
                '--margins-bottom={0}'.format(marginconf[2]),
                '--margins-left={0}'.format(marginconf[3]),
            ])
        if self._enhanced:
            generation_cmd.append(unicode(img_dir))
        else:
            generation_cmd.extend([unicode(x)
                                   for x in sorted(img_dir.iterdir())])
        generation_cmd.append(unicode(out_dir))
        logger.debug(" ".join(generation_cmd))
        subprocess.call(generation_cmd)

    def _split_configuration(self, projectfile, temp_dir):
        num_pieces = multiprocessing.cpu_count()
        tree = ET(file=unicode(projectfile))
        num_files = len(tree.findall('./files/file'))
        splitfiles = []
        files_per_job = int(math.ceil(float(num_files)/num_pieces))
        for idx in xrange(num_pieces):
            tree = ET(file=unicode(projectfile))
            root = tree.getroot()
            start = idx*files_per_job
            end = start + files_per_job
            if end > num_files:
                end = None
            for elem in ('files', 'images', 'pages',
                         'file-name-disambiguation'):
                elem_root = root.find(elem)
                to_keep = elem_root.getchildren()[start:end]
                to_remove = [x for x in elem_root.getchildren()
                             if not x in to_keep]
                for node in to_remove:
                    elem_root.remove(node)
            out_file = temp_dir / "{0}-{1}.ScanTailor".format(projectfile.stem,
                                                              idx)
            tree.write(unicode(out_file))
            splitfiles.append(out_file)
        return splitfiles

    def _generate_output(self, projectfile, out_dir):
        logger.debug("Generating output...")
        temp_dir = Path(tempfile.mkdtemp(prefix="spreads."))
        split_config = self._split_configuration(projectfile, temp_dir)
        logger.debug("Launching those subprocesses!")
        processes = [subprocess.Popen(['scantailor-cli', '--start-filter=6',
                                       unicode(cfgfile), unicode(out_dir)])
                     for cfgfile in split_config]
        while processes:
            for p in processes[:]:
                if p.poll() is not None:
                    processes.remove(p)
        shutil.rmtree(unicode(temp_dir))

    def process(self, path):
        autopilot = self.config['autopilot'].get(bool)
        if not autopilot and not find_in_path('scantailor'):
            raise MissingDependencyException(
                "Could not find executable `scantailor` in"
                " $PATH. Please install the appropriate"
                " package(s)!")
        projectfile = path / "{0}.ScanTailor".format(path.name)
        img_dir = path / 'raw'
        out_dir = path / 'done'

        if not projectfile.exists():
            self._generate_configuration(projectfile, img_dir, out_dir)
        if not autopilot:
            logger.info("Opening ScanTailor GUI for manual adjustment")
            subprocess.call(['scantailor', unicode(projectfile)])
        logger.info("Generating output images from ScanTailor configuration.")
        self._generate_output(projectfile, out_dir)



import unittest

import spreads.vendor.confit as confit
from PySide.QtGui import QPixmap, QImage, QApplication
from mock import patch, MagicMock as Mock
from spreads.vendor.pathlib import Path

import spreads.plugin as plugin
import spreads.workflow as workflow
import spreadsplug.gui.gui as gui
with patch('subprocess.check_output'):
    import spreadsplug.tesseract as tess


class TestWizard(unittest.TestCase):
    def setUp(self):
        try:
            self.app = QApplication([])
        except RuntimeError:
            # TODO: Somehow our app gets created multiple times, so we just
            #       ignore that here...
            pass
        self.config = confit.Configuration('test_gui')
        self.config['plugins'] = [u'tesseract', u'scantailor']
        self.config['driver'] = u'dummy'
        self.workflow = workflow.Workflow(config=self.config,
                                          path='/tmp/foobar')
        plugin.setup_plugin_config(self.config)

        tess.AVAILABLE_LANGS = ["en"]
        # TODO: Cams ought to be 'odd' and 'even'!
        self.cams = [Mock(), Mock()]
        workflow.get_devices = Mock(return_value=self.cams)
        # Mock out message boxes
        gui.QtGui.QMessageBox = Mock()
        gui.QtGui.QMessageBox.exec_.return_value = True
        self.wizard = gui.SpreadsWizard(self.config)
        self.wizard.show()

    def tearDown(self):
        gui.QtGui.QImage = QImage
        gui.QtGui.QPixmap = QPixmap

    def test_intro_page(self):
        page = self.wizard.page(0)
        page.initializePage()
        page.line_edit.setText('/tmp/foobar')
        assert page.validatePage()
        assert self.wizard.workflow.path == Path('/tmp/foobar')
        #assert spreads.config['first_page'].get(unicode) == "left"
        #assert not spreads.config['rotate_inverse'].get(bool)
        #assert not spreads.config['autopilot'].get(bool)
        assert (self.config['scantailor']['detection'].get(unicode)
                == u"content")
        #assert spreads.config['language'].get(str) == 'eng'

        # TODO: Check option boxes
        # TODO: Select path
        # TODO: Assert that spreads.config is updated accordingly
        # TODO: Assert that the project path is set correctly

    def test_intro_page_nopath(self):
        page = self.wizard.page(0)
        page.initializePage()
        assert not page.validatePage()

    def test_capture_page(self):
        self.wizard.workflow = self.workflow
        self.wizard.selected_devices = self.cams
        gui.QtGui.QImage = Mock()
        gui.QtGui.QPixmap.fromImage = Mock(return_value=QPixmap())
        page = self.wizard.page(1)
        page.initializePage()
        # TODO: Test capture triggering, logbox updates, etc
        assert page.validatePage()

    def test_postprocess_page(self):
        self.wizard.workflow = self.workflow
        page = self.wizard.page(3)
        page.initializePage()
        # TODO: See that logbox works, postprocess is executed
        assert page.validatePage()



import unittest

import pytest
import spreads.vendor.confit as confit
from mock import MagicMock as Mock, patch, call

import spreads.util as util
import spreadsplug.dev.chdkcamera as chdkcamera

util.find_in_path = Mock(return_value=True)


class TestChdkCameraDevice(unittest.TestCase):
    def setUp(self):
        chdkcamera.subprocess.check_output = Mock()
        usbmock = Mock()
        usbmock.bus, usbmock.address = 1, 2
        self.config = confit.Configuration('test_chdkcamera')
        tmpl = chdkcamera.CHDKCameraDevice.configuration_template()
        for key, option in tmpl.items():
            if option.selectable:
                self.config[key] = option.value[0]
            else:
                self.config[key] = option.value
        self.config['chdkptp_path'] = u'/tmp/chdkptp'
        self.config['two_devices'] = False
        with patch('spreadsplug.dev.chdkcamera.'
                   'CHDKCameraDevice._execute_lua') as lua:
            lua.return_value = {'build_revision': 3000}
            self.dev = chdkcamera.CHDKCameraDevice(self.config, usbmock)

    def test_init_noremote(self):
        usbmock = Mock()
        usbmock.bus, usbmock.address = 1, 2
        with patch('spreadsplug.dev.chdkcamera.'
                   'CHDKCameraDevice._execute_lua') as lua:
            lua.return_value = {'build_revision': 2500}
            dev = chdkcamera.CHDKCameraDevice(self.config, usbmock)
            assert dev._can_remote is False

    def test_run(self):
        chdkcamera.subprocess.check_output.return_value = (
            "connected: foo bar \n"
            "asdf")
        output = self.dev._run('foobar')
        assert output == ["asdf"]
        assert (call([u'/tmp/chdkptp/chdkptp', '-c-d=002 -b=001',
                      '-eset cli_verbose=2', '-efoobar'],
                     env={'LUA_PATH': u'/tmp/chdkptp/lua/?.lua'}, stderr=-2)
                in chdkcamera.subprocess.check_output.call_args_list)

    def test_execute_lua(self):
        self.dev._run = Mock()
        self.dev._execute_lua("foobar")
        self.dev._run.assert_called_with("luar foobar")

    def test_execute_lua_nowait(self):
        self.dev._run = Mock()
        self.dev._execute_lua("foobar", wait=False)
        self.dev._run.assert_called_with("lua foobar")

    def test_execute_lua_with_get_result(self):
        self.dev._run = Mock()
        self.dev._run.return_value = ["10:return:'bar'"]
        assert self.dev._execute_lua("foo", get_result=True) == "bar"

    def test_execute_lua_with_get_result_hasreturn(self):
        self.dev._run = Mock()
        self.dev._run.return_value = ["10:return:5"]
        out = self.dev._execute_lua("return moo", get_result=True)
        assert out == 5
        self.dev._run.assert_called_with("luar return moo")

    def test_get_target_page(self):
        self.dev._run = Mock()
        with patch('spreadsplug.dev.chdkcamera.open',
                   create=True) as mock_open:
            mock_open.return_value = Mock(spec=file)
            fhandle = mock_open.return_value.__enter__.return_value
            fhandle.readline.return_value = 'ODD'
            target_page = self.dev._get_target_page()
        assert target_page == 'odd'

    def test_get_target_page_error(self):
        self.dev._run = Mock()
        self.dev._run.side_effect = [util.DeviceException]
        with pytest.raises(ValueError) as excinfo:
            self.dev._get_target_page()

    @patch('spreadsplug.dev.chdkcamera.os.write')
    def test_set_target_page(self, write):
        self.dev._run = Mock()
        self.dev.set_target_page('odd')
        assert self.dev.target_page == 'odd'

    def test_get_preview_image(self):
        self.dev._run = Mock()
        with patch('spreadsplug.dev.chdkcamera.open',
                   create=True) as mock_open:
            fhandle = mock_open.return_value.__enter__.return_value
            fhandle.read.return_value = 'foobar'
            data = self.dev.get_preview_image()
        assert data == 'foobar'
